#!/usr/bin/perl
use strict;
use Socket;		# socket getprotobyname connect ...
use IO::Handle;		# autoflush

$|=1; # not 0 - then flush after every print/write

my $ttynum = shift || 2;	# which tty to tap
my $wide = 80;		# width TTY
my $high = 25;		# height TTY
my $lcdw = 20;		# width LCD
my $lcdh = 4;		# height LCD
my $xpos = 0;
my $ypos = 0;
my $inhalt;		# tty contents
my $y;			# counter
my $key;		# most recently pressed key

unless ($ttynum =~ /^[0-9][0-9]?$/) {
	die "Argument must be a TTY number";
}

socket(RAUS, PF_INET, SOCK_STREAM, getprotobyname('tcp')) || die "socket: $!";
connect(RAUS, sockaddr_in(13666, inet_aton('localhost'))) || die "no LCDd loaded?";
RAUS->autoflush(1);			# make RAUS unbuffered
# similar to: select(RAUS); $|=1;	# make RAUS unbuffered

print "Use cursor keys or u d r l to view parts of TTY$ttynum on the LCD\n";
print "To quit, press q or ctrl-c\n";

print RAUS "hello\n";	# could parse LCD size from reply
$_ = <RAUS>;
# "connect LCDproc 0.x.x protocol 0.3 lcd wid 20 hgt 4 cellwid 5 cellhgt 8"
print RAUS "screen_add tap\n";
$_ = <RAUS>;
# "success"
print RAUS "screen_set tap -priority 64 -name foo -wid 20 -hgt 4 -heartbeat off\n";
$_ = <RAUS>;
# "listen tap"

for($y=1; $y<=$lcdh; $y++) {
	print RAUS "widget_add tap line$y string\n";
	$_ = <RAUS>;
	print RAUS "widget_set tap line$y 1 $y \"line $y ready\"\n";
	$_ = <RAUS>;
}

# make ctrl-c / ctrl-d a normal key, and suppress echoing back:
system "stty", 'raw', '-echo';	# by eric

# according to man perlfunc, getc would otherwise wait for "enter"...
# man perlfunc suggests: system "stty", 'icanon', 'eol', '\001';
# ... or use POSIX::getattr() in some way? or Term::ReadKey?
# or stty igncr / stty -igncr? or -cooked?

$key = ' ';
while(1) {

	my $readvec = '';
	my ($ready, $timeleft);
	vec($readvec,fileno(STDIN),1) = 1;	# only check for 1 handle
	($ready, $timeleft) = 
		select $readvec, undef, undef, 0.2;	# uses the select(2) syscall
		# check for pending input - not for "can write" and
		# not for "exception pending". Wait max 0.2 seconds.
	undef $key;
	if ($ready) {
		sysread STDIN, $key, 1;		# get one char, unbuffered!
	} else {
		$key = ' ';			# not EOF, just no key...
	}

	unless (open(REIN, "< /dev/vcs$ttynum")) {
		# $inhalt = "could not read screen contents";
		$inhalt = "";
		for ($y=1; $y<=$high; $y++) {
			$inhalt .= substr("dummy line $y " x 100, 0, $wide);
		}
	} else {
		if (open(FULL, "< /dev/vcsa$ttynum")) {
			my $coords;
			sysread FULL, $coords, 4;
			# vcsa is bytes: rows cols ?? ??
			# and then char, attrib, char, attrib...
			$high = unpack('c',substr($coords,0,1));
			$wide = unpack('c',substr($coords,1,1));
			$xpos = ($wide-$lcdw) if ($xpos > ($wide-$lcdw));
			$ypos = ($high-$lcdh) if ($ypos > ($high-$lcdh));
			close(FULL);
		} # figure out current screen size
		read REIN, $inhalt, $wide * $high;
		close(REIN);
		# print ".";
	} # fetch current screen contents

	if ((defined $key) && ($key ne 'q') && ($key ne "\003")) {
		# in combination with getc(), test for (defined $key) here
		if ($key eq "\033") {
			$key = getc(STDIN);
			# no check for definedness here
			if ($key eq '[') {	# ESC [ ... sequence?
				$key = getc(STDIN);
				$key =~ tr/ABCD/udrl/;
			}
		}
		if (($key eq 'l') || ($key eq "\033[D")) {
			$xpos-- unless ($xpos == 0);
		} elsif ($key eq 'r') { 
			$xpos++ unless ($xpos == ($wide-$lcdw));
		} elsif ($key eq 'u') {
			$ypos-- unless ($ypos == 0);
		} elsif ($key eq 'd') {
			$ypos++ unless ($ypos == ($high-$lcdh));
		} elsif ($key eq 'h') {
			$xpos=0; $ypos=0;
		}
	} else {
		print RAUS "screen_del tap\n";
		$_ = <RAUS>;
		close(RAUS);
		system "stty", 'cooked', 'echo';
		print "done\n";
		# according to man perlfunc, getc would otherwise wait for "enter"...
		# system "stty", 'icanon', 'eol', '^@';	# ASCII null
		exit();
	} # exit

	for($y=1; $y<=$lcdh; $y++) {
		my $cols = ($wide - $xpos);
		$cols = $lcdw if ($cols > $lcdw);
		my $offset = ($wide * ($ypos + ($y-1))) + $xpos;
		my $text = substr($inhalt, $offset, $cols);
		print RAUS "widget_set tap line$y 1 $y \"$text\"\n";
		$_ = <RAUS>;
	} # LCD update loop
	### $key = getc(STDIN);

} # infinite loop
