	; public domain "LCD Braille" tool by Eric Auer 2004
	; shows a part of the DOS screen on a winamp-wired 20x4 LCD on LPT1
	; display. To navigate, use keys connected to LPT1 like:
	; data bit -!<!- key - nACK line (optionally add pull-up to nACK)

	; Buttons: bit 0..7 = blue yellow red up right down left center
	; Wiring: bit 0..7 = LCD pin 7..14, nSTB = LCD pin 6 (EN)
	; (EN is: 1 read, 0 tristate, 1->0 write to LCD. Hardware inverts STB!)
	; nINIT = LCD pin 4 (RS: 1 data, 0 command)
	; AutoFeed = LCD pin 5 (RW: 1 R, 0 W to LCD. Hardware inverts!)
	; 1 - 2 + (5 Volts) 3 contrast (typ. close to 0 Volts)
	; 15 LED+ (connect through ca. 50 Ohms to +) 16 LED- (connect to -)

	org 0x100
start:	push es
	mov es,[cs:0x2c]	; environment segment
	mov ah,0x49	; free (de-alloc)
	int 0x21
	pop es

	push ds
	xor ax,ax
	mov ds,ax
	mov ax,[ds:0x408]	; port of LPT1
	pop ds
	mov [cs:port],ax

	mov dx,[cs:port]
	add dx,2
	mov al,0xcb	; dont start yet
	out dx,al

	mov al,0x38	; 8 bit mode... 011b 10?? (busy flag not yet valid?)
	call cmd
	mov al,0x38	; second attempt: 8 bit mode... 011b 10??
	call cmd
	mov al,3	; clear/home 0000 0010
	call cmd
	mov al,6	; no auto-scroll, but auto-address-inc
	call cmd

	push es
	mov ax,0x351c	; get timer tick handler
	int 0x21
	mov [old1c],bx
	mov [old1c+2],es
	pop es
	mov ax,0x251c	; set timer tick handler
	mov dx,new1c	; our handler
	push cs
	pop ds
	int 0x21

	mov ah,9
	mov dx,hellomsg
	int 0x21
	mov dx,lastbyte+15
	mov cl,4
	shr dx,cl
	mov ax,0x3100	; go TSR
	int 0x21

; ------------------

cmd:	mov ah,0xca	; topic: command (send COMMAND to LCD)
cmddat:
	push dx
	push ax
	mov dx,[cs:port]
	call waitlcd
	out dx,al
	inc dx
	inc dx
	mov al,ah
	out dx,al	; data send strobe is falling edge...
	or al,1
	push ax
	in al,dx	; waste a bit of time
	pop ax
	out dx,al
	pop ax
	pop dx
	ret

dat:	mov ah,0xce	; topic: data (send DATA to LCD)
	jmp cmddat

waitlcd:	push dx
	push ax
	mov dx,[cs:port]
	inc dx
	inc dx
	mov al,0xe8	; input direction, read command register (not data register)
	out dx,al
	dec dx
	dec dx
waithere:
	in al,dx	; highest bit of both registers is busy flag, not real data
	test al,0x80	; busy?
	jnz waithere	; then wait
	pop ax
	pop dx
	ret

; ------------------

getkeys:	; collect button states in AL, return NZ if some pressed
	push dx
	xor ax,ax
	mov ah,1	; bit to test
	mov dx,[cs:port]

keyloop:
	xchg al,ah
	not al	; inverse! set all bits except to-be-tested one
	out dx,al
	not al	; normal again
	xchg al,ah

	push ax
	inc dx
	in al,dx	; status port
	dec dx
	test al,0x40	; will be zero if that button pressed
	pop ax

	jnz notkey
	or al,ah	; set that bit

notkey:
	shl ah,1	; on to the next bit
	or ah,ah	; all done?
	jnz keyloop
	or al,al	; test if zero
	pop dx
	ret

; ------------------

new1c:	push ax	; new int 1c (timer tick) handler
	push ds
	xor ax,ax
	mov ds,ax
	mov ax,[ds:0x46c]	; current timer tick
	test ax,1	; odd?
	pop ds
	jz goodtick	; handle even ticks
	jmp badtick	; do not handle odd ticks (reduces load + "button speed")
	; TODO: reduce speed even further for the blue/yellow key

goodtick:
	push bx
	push cx
	mov bx,[cs:xcoord]
	mov cx,[cs:ycoord]

	call getkeys
	jz nokeys

%if 0	; debugging: show key info ("push word" needs non-ancient CPU)
	push ds
	push word 0xb800
	pop ds
	mov ah,0x1e
	mov [ds:0],ax
	pop ds
%endif

	test al,1	; blue key: 20 chars towards upper left
	jz not1
	sub bx,20	; X
	or bx,bx
	jns not1
	or cx,cx	; Y already at top?
	jz short ishome
	dec cx	; Y
	mov bx,[cs:xmax]	; wrap to 1 line above, right end

not1:	test al,2	; yellow key: 20 chars towards bottom right
	jz not2
	add bx,20	; X
	cmp bx,[cs:xmax]
	jbe not2
	cmp cx,[cs:ymax]	; Y already at bottom?
	jae isend
	inc cx	; Y
	mov bx,0	; wrap to 1 line below, left end
	jmp short not2
isend:	mov bx,[cs:xmax]

not2:	test al,4	; red key: home
	jz not4

ishome:	xor bx,bx
	xor cx,cx

not4:	test al,8	; UP
	jz not8
	or cx,cx	; Y
	jz not8
	dec cx

not8:	test al,0x10	; RIGHT
	jz not10
	cmp bx,[cs:xmax]
	jae not10
	inc bx

not10:	test al,0x20	; DOWN
	jz not20
	cmp cx,[cs:ymax]
	jae not20
	inc cx

not20:	test al,0x40	; LEFT
	jz not40
	or bx,bx
	jz not40
	dec bx

not40:	test al,0x80	; center key: freeze LCD contents (toggle)
	jz not80
	not byte [cs:frozen]

not80:

	mov [cs:xcoord],bx
	mov [cs:ycoord],cx

nokeys:
	mov al,[cs:frozen]	; flag to block LCD updates
	test al,al
	jnz short noshow	; skip LCD updates

	push ds
	mov ax,0xb800	; TODO: allow other than color text mode, too
	mov ds,ax
	xor bx,bx	; LCD X
	xor cx,cx	; LCD Y

onechar:
	push si
	mov si,bx	; X
	add si,[cs:xcoord]
	add si,si	; convert to framebuffer offset
	mov ax,cx	; Y
	add ax,[cs:ycoord]
	push dx
	mov dx,80*2	; TODO: adjust to current mode
	mul dx
	pop dx
	add si,ax	; framebuffer offset is complete now
	mov ax,[ds:si]	; AL is char. TODO: process AH, color, too
	; TODO: process screen cursor position to show cursor on LCD

	mov si,cx	; LCD row
	add si,si
	push ax
	mov ax,[cs:rowoffsets+si]
	add ax,bx	; LCD column
	add ax,0x80	; command: select address
	call cmd	; send command
	pop ax
	call dat	; send char. TODO: translate 0..1f / 80..9f / a0..df / e0..ff ?
	pop si

	inc bx	; LCD column
	cmp bx,20
	jb onechar
	xor bx,bx	; LCD column
	inc cx	; LCD row
	cmp cx,4
	jb onechar
	pop ds

noshow:
	pop cx
	pop bx

badtick:
	pop ax
chain	db 0xea	; jmp far immediate... to pointer old1c
old1c	dw 0,0	; old int 1c handler address

; ------------------

port	dw 0x378	; printer port
rowoffsets	dw 0, 64, 20, 84	; LCD text row offsets in LCD memory
xcoord	dw 0	; screen x coordinate of LCD left edge
ycoord	dw 0	; screen y coordinate of LCD top
xmax	dw 80-20	; TODO: adjust for screen width
ymax	dw 25-4	; TODO: adjust for screen height
frozen	db 0	; set if LCD contents have to stay unchanged

	; unused: command 01cccrrr sets row r of user defined char c
	;   to data (5 low bits used, rightmost is lsb)
	; charset: 0..7 userdef, (8..F repeat of userdef), (10..1F empty)
	;   20..7F ASCII, (80..9F empty), A0..DF japanese, E0..FF greek etc.
lastbyte:	; end of TSR part

; ------------------

hellomsg	db "Eric's TSR BrlTTYish LCD driver loaded",13,10
	db "This is free (public domain) software by Eric Auer 9/2004",13,10
	db "$"

